@@ -12,10 +12,11 @@ import io
 import re
 import socket
 
+from . import keys
 from .compat import text_type
 
 
-__version__ = '0.3'
+__version__ = '0.4'
 __all__ = ['notify', 'Notifier', 'RawIcon', 'SocketCallback', 'Event']
 
 
@@ -76,6 +77,8 @@ class Notifier(object):
         self.name = name
         self.icon = icon
         self.events = coerce_to_events(event_defs)
+        if not self.events:
+            raise GNTPError('You have to set at least one notification type')
         if gntp_client_class is None:
             gntp_client_class = GNTPClient
         self.gntp_client = gntp_client_class(**kwargs)
@@ -87,7 +90,7 @@ class Notifier(object):
                          the GNTP server.  Defaults to `None`.
         """
         request = RegisterRequest(self.name, self.icon, self.events)
-        self.gntp_client.register(request, callback)
+        self.gntp_client.process_request(request, callback)
 
     def notify(self, name, title, text='', id_=None, sticky=False,
                priority=0, icon=None, coalescing_id=None, callback=None,
@@ -125,6 +128,9 @@ class Notifier(object):
 
         .. note:: In Growl 1.3.3, `icon` of url string does not work.
 
+        .. note:: Growl for Windows v2.0+ and Growl v1.3+ require
+                  `coalescing_id` to be the same on both the original and
+                  updated notifcation, ignoring the value of `id_`.
         """
         notification = Notification(name, title, text, id_=id_, sticky=sticky,
                                     priority=priority, icon=icon,
@@ -132,8 +138,8 @@ class Notifier(object):
                                     gntp_callback=gntp_callback,
                                     **socket_callback_options)
         request = NotifyRequest(self.name, notification)
-        self.gntp_client.notify(request, callback,
-                                notification.socket_callback)
+        self.gntp_client.process_request(
+            request, callback, socket_callback=notification.socket_callback)
 
 
 class GNTPError(Exception):
@@ -186,40 +192,7 @@ class Notification(object):
             return self.callback
 
 
-class BaseRequest(object):
-    """Abstract base class for GNTP request."""
-    #: Request message type.  Subclasses must override this attribute.
-    #: Currently ``REGISTER`` and ``NOTIFY`` are supported.
-    message_type = None
-
-    @property
-    def message(self):
-        """Return utf-8 encoded request message."""
-        return b''.join([b'GNTP/1.0 ',
-                         self.message_type.encode('utf-8'),
-                         b' NONE',
-                         LINE_DELIMITER,
-                         self.message_body,
-                         LINE_DELIMITER])
-
-    @property
-    def message_body(self):
-        """Return utf-8 encoded message body."""
-        writer = io.BytesIO()
-        header_writer = HeaderWriter(writer)
-        section_writer = SectionWriter(writer, header_writer)
-        self._write_into(header_writer)
-        self._write_into(section_writer)
-        return writer.getvalue()
-
-    def _write_into(self, writer):
-        """Subclasses must override this method to serialize their message
-        bodies.
-        """
-        raise NotImplementedError
-
-
-class RegisterRequest(BaseRequest):
+class RegisterRequest(object):
     """Represent ``REGISTER`` request.
 
     :param app_name: the name of the application.
@@ -236,11 +209,11 @@ class RegisterRequest(BaseRequest):
         self.app_icon = coerce_to_icon(app_icon)
         self.events = events
 
-    def _write_into(self, writer):
+    def write_into(self, writer):
         writer.write_register_request(self)
 
 
-class NotifyRequest(BaseRequest):
+class NotifyRequest(object):
     """Represent ``NOTIFY`` request.
 
     :param app_name: the name of the application.
@@ -252,7 +225,7 @@ class NotifyRequest(BaseRequest):
         self.app_name = app_name
         self.notification = notification
 
-    def _write_into(self, writer):
+    def write_into(self, writer):
         writer.write_notify_request(self)
 
 
@@ -269,27 +242,17 @@ class Response(object):
         self.headers = headers
 
 
-class GNTPConnectionCallbacksMixin(object):
-    """Define callbacks run by the GNTP connection."""
+class BaseGNTPConnection(object):
+    """Abstract base class for GNTP connection."""
 
-    def on_register_message(self, message):
-        r"""Callback for ``-OK`` response of ``REGISTER`` request.
+    def __init__(self, final_callback, socket_callback=None):
+        self.final_callback = final_callback
+        self.socket_callback = socket_callback
 
-        :param message: string of response terminated by
-                        `'\\r\\n\\r\\n\\r\\n'`.
-        """
-        try:
-            response = parse_response(message, '-OK')
-        finally:
-            self.close()
-        if self.final_callback is not None:
-            self.final_callback(response)
+    def on_ok_message(self, message):
+        r"""Callback for ``-OK`` response.
 
-    def on_notify_message(self, message):
-        r"""Callback for ``-OK`` response of ``NOTIFY`` request.
-
-        :param message: string of response terminated by
-                        `'\\r\\n\\r\\n\\r\\n'`.
+        :param message: string of response terminated by `'\\r\\n\\r\\n'`.
         """
         try:
             response = parse_response(message, '-OK')
@@ -304,29 +267,42 @@ class GNTPConnectionCallbacksMixin(object):
     def on_callback_message(self, message):
         r"""Callback for ``-CALLBACK`` response.
 
-        :param message: string of response terminated by
-                        `'\\r\\n\\r\\n\\r\\n'`.
+        :param message: string of response terminated by `'\\r\\n\\r\\n'`.
         """
         try:
-            callback_response = parse_response(message, '-CALLBACK')
-            response = self.socket_callback(callback_response)
+            response = parse_response(message, '-CALLBACK')
+            callback_result = self.socket_callback(response)
         finally:
             self.close()
         if self.final_callback is not None:
-            self.final_callback(response)
+            self.final_callback(callback_result)
+
+    def write_message(self, message):
+        """Subclasses must override this method to send a message to the GNTP
+        server."""
+        raise NotImplementedError
+
+    def read_message(self, callback):
+        """Subclasses must override this method to receive a message from the
+        GNTP server."""
+        raise NotImplementedError
+
+    def close(self):
+        """Subclasses must override this method to close the connection with
+        the GNTP server."""
+        raise NotImplementedError
 
 
-class GNTPConnection(GNTPConnectionCallbacksMixin):
+class GNTPConnection(BaseGNTPConnection):
     """Represent the connection with the GNTP server."""
 
     def __init__(self, address, timeout, final_callback, socket_callback=None):
+        BaseGNTPConnection.__init__(self, final_callback, socket_callback)
         self.sock = socket.create_connection(address, timeout=timeout)
-        self.final_callback = final_callback
-        self.socket_callback = socket_callback
 
-    def write_message(self, request):
+    def write_message(self, message):
         """Send the request message to the GNTP server."""
-        self.sock.send(request.message)
+        self.sock.send(message)
 
     def read_message(self, callback):
         """Read a message from opened socket and run callback with it."""
@@ -340,38 +316,49 @@ class GNTPConnection(GNTPConnectionCallbacksMixin):
 
 
 class GNTPClient(object):
-    """Tiny GNTP client
+    """GNTP client.
 
-    :param host: host of GNTP server.  Defaults to ``'localhost'``.
-    :param port: port of GNTP server.  Defaults to ``23053``.
-    :param timeout: timeout.  Defaults to ``10``.
+    :param host: host of GNTP server.  Defaults to `'localhost'`.
+    :param port: port of GNTP server.  Defaults to `23053`.
+    :param timeout: timeout in seconds.  Defaults to `10`.
+    :param password: the password used in creating the key.
+    :param key_hashing: the type of hash algorithm used in creating the key.
+                        It is `keys.MD5`, `keys.SHA1`, `keys.SHA256` or
+                        `keys.SHA512`.  Defaults to `keys.SHA256`.
+    :param encryption: the tyep of encryption algorithm used.
+                       It is `None`, `ciphers.AES`, `ciphers.DES` or
+                       `ciphers.3DES`.  `None` means no encryption.
+                       Defaults to `None`.
+    :param connection_class: GNTP connection class.  If it is `None`,
+                             :class:`GNTPConnection` is used.  Defaults to
+                             `None`.
     """
 
-    connection_class = GNTPConnection
-
-    def __init__(self, host='localhost', port=23053, timeout=10):
+    def __init__(self, host='localhost', port=23053, timeout=10,
+                 password=None, key_hashing=keys.SHA256, encryption=None,
+                 connection_class=None):
         self.address = (host, port)
         self.timeout = timeout
+        self.connection_class = connection_class or GNTPConnection
+        if (encryption is not None and
+            encryption.key_size > key_hashing.key_size):
+            raise GNTPError('key_hashing key size (%s:%d) must be at'
+                            ' least encryption key size (%s:%d)' % (
+                    key_hashing.algorithm_id, key_hashing.key_size,
+                    encryption.algorithm_id, encryption.key_size))
+        self.packer_factory = RequestPackerFactory(password, key_hashing,
+                                                   encryption)
 
-    def register(self, request, callback):
-        """Open a connection and process the register request.
+    def process_request(self, request, callback, **kwargs):
+        """Process a request.
 
         :param callback: the final callback run after closing connection.
         """
-        conn = self._connect(callback)
-        conn.write_message(request)
-        conn.read_message(conn.on_register_message)
-
-    def notify(self, request, callback, socket_callback):
-        """Open a connection and process the notify request.
-
-        :param callback: the final callback run after closing connection.
-        :param socket_callback: :class:`SocketCallback` instance called by
-                                `callback_message_callback`.
-        """
-        conn = self._connect(callback, socket_callback=socket_callback)
-        conn.write_message(request)
-        conn.read_message(conn.on_notify_message)
+        packer = self.packer_factory.create()
+        message = packer.pack(request)
+        conn = self._connect(callback, **kwargs)
+        conn.write_message(message)
+        conn.read_message(conn.on_ok_message)
 
     def _connect(self, final_callback, **kwargs):
         """Connect to the GNTP server and return the connection."""
@@ -386,7 +373,6 @@ def generate_messages(sock, size=1024):
         buf += sock.recv(size)
         if not buf:
             break
-        # or .find(LINE_DELIMITER + MESSAGE_DELIMITER)?
         pos = buf.find(MESSAGE_DELIMITER)
         if ((pos < 0 and len(buf) >= MAX_MESSAGE_SIZE) or
             (pos > MAX_MESSAGE_SIZE - 4)):
@@ -400,7 +386,7 @@ def generate_messages(sock, size=1024):
 def parse_response(message, expected_message_type=None):
     """Parse response and return response object."""
     try:
-        lines = message.split(LINE_DELIMITER)
+        lines = [line for line in message.split(LINE_DELIMITER) if line]
         _, message_type = parse_information_line(lines.pop(0))
         if (expected_message_type is not None and
             expected_message_type != message_type):
@@ -408,7 +394,7 @@ def parse_response(message, expected_message_type=None):
                     message_type, expected_message_type))
 
         headers = dict([s.strip().decode('utf-8') for s in line.split(b':', 1)]
-                       for line in lines if line)
+                       for line in lines)
         if message_type == '-ERROR':
             raise GNTPError('%s: %s' % (headers['Error-Code'],
                                         headers['Error-Description']))
@@ -515,8 +501,9 @@ class SocketCallback(object):
     :class:`Response` instance.
 
     :param context: value of ``Notification-Callback-Context``.
+                    Defaults to ``'None'``.
     :param context-type: value of ``Notification-Callback-Context-Type``.
-                         Defaults to ``'text/plain; charset=utf-8'``.
+                         Defaults to ``'None'``.
     :param on_click: the callback run at ``CLICKED`` callback result.
     :param on_close: the callback run at ``CLOSED`` callback result.
     :param on_timeout: the callback run at ``TIMEDOUT`` callback result.
@@ -524,7 +511,7 @@ class SocketCallback(object):
     .. note:: TIMEDOUT callback does not occur in my Growl 1.3.3.
     """
 
-    def __init__(self, context='', context_type='text/plain; charset=utf-8',
+    def __init__(self, context='None', context_type='None',
                  on_click=None, on_close=None, on_timeout=None):
         self.context = context
         self.context_type = context_type
@@ -601,6 +588,96 @@ def coerce_to_callback(gntp_callback=None, **socket_callback_options):
         return SocketCallback(**socket_callback_options)
 
 
+class _NullCipher(object):
+    """Null object for the encryption of messages."""
+
+    algorithm = None
+    algorithm_id = 'NONE'
+    encrypt = lambda self, text: text
+    decrypt = lambda self, text: text
+    __bool__ = lambda self: False
+    __nonzero__ = __bool__
+
+
+NullCipher = _NullCipher()
+
+
+class RequestPackerFactory(object):
+    """The factory of :class:`RequestPacker`.
+
+    If `password` is None, `hashing` and `encryption` are ignored.
+    """
+
+    def __init__(self, password=None, hashing=keys.SHA256, encryption=None):
+        self.password = password
+        self.hashing = password and hashing
+        self.encryption = (password and encryption) or NullCipher
+
+    def create(self):
+        """Create an instance of :class:`RequestPacker` and return it."""
+        key = self.password and self.hashing.key(self.password)
+        cipher = self.encryption and self.encryption.cipher(key)
+        return RequestPacker(key, cipher)
+
+
+class RequestPacker(object):
+    """The serializer of requests.
+
+    `key` and `cipher` have random-generated salt and iv respectively.
+
+    :param key: an instance of :class:`keys.Key`.
+    :param cipher: an instance of :class:`ciphers.Cipher` or `NullCipher`.
+    """
+
+    def __init__(self, key=None, cipher=None):
+        self.key = key
+        self.cipher = cipher or NullCipher
+
+    def pack(self, request):
+        """Return utf-8 encoded request message."""
+        return (self._information_line(request) +
+                LINE_DELIMITER +
+                self._headers(request) +
+                self._sections(request) +
+                LINE_DELIMITER)
+
+    def _information_line(self, request):
+        """Return utf-8 encoded information line."""
+        result = (b'GNTP/1.0 ' +
+                  request.message_type.encode('utf-8') +
+                  b' ' +
+                  self.cipher.algorithm_id.encode('utf-8'))
+        if self.cipher.algorithm is not None:
+            result += b':' + self.cipher.iv_hex
+        if self.key is not None:
+            result += (b' ' +
+                       self.key.algorithm_id.encode('utf-8') +
+                       b':' +
+                       self.key.key_hash_hex +
+                       b'.' +
+                       self.key.salt_hex)
+        return result
+
+    def _headers(self, request):
+        """Return utf-8 encoded headers."""
+        writer = io.BytesIO()
+        header_writer = HeaderWriter(writer)
+        request.write_into(header_writer)
+        headers = writer.getvalue()
+        result = self.cipher.encrypt(headers)
+        if self.cipher.algorithm is not None:
+            result += LINE_DELIMITER
+        return result
+
+    def _sections(self, request):
+        """Return utf-8 encoded message body."""
+        writer = io.BytesIO()
+        header_writer = HeaderWriter(writer)
+        section_writer = SectionWriter(writer, header_writer, self.cipher)
+        request.write_into(section_writer)
+        return writer.getvalue()
+
+
 class HeaderWriter(object):
 
     def __init__(self, writer):
@@ -651,9 +728,8 @@ class HeaderWriter(object):
     def write(self, name, value):
         """Write utf-8 encoded headers into writer.
 
-        `headers` is iterable of (`header name`, `header value`) tuple.
-        If `header value` is instance of unicode, it is encoded.
-        Otherwise it is converted with ``str`` function.
+        :param name: the name of the header.
+        :param value: the value of the header.
         """
         if value is not None:
             if not isinstance(value, bytes):
@@ -666,9 +742,10 @@ class HeaderWriter(object):
 
 class SectionWriter(object):
 
-    def __init__(self, writer, header_writer):
+    def __init__(self, writer, header_writer, cipher):
         self.writer = writer
         self.header_writer = header_writer
+        self.cipher = cipher
 
     def write_register_request(self, request):
         self.write_icon(request.app_icon)
@@ -685,9 +762,10 @@ class SectionWriter(object):
         self.write_icon(notification.icon)
 
     def write_raw_icon(self, icon):
+        icon_data = self.cipher.encrypt(icon.data)
         headers = [(b'Identifier', icon.unique_id()),
-                   (b'Length', len(icon.data))]
-        self.write(headers, icon.data)
+                   (b'Length', len(icon_data))]
+        self.write(headers, icon_data)
 
     def write_icon(self, icon):
         if icon is not None:
@@ -696,14 +774,12 @@ class SectionWriter(object):
     def write(self, headers, body):
         """Write utf-8 encoded sections into writer.
 
-        `headers` is iterable of (`section headers`, `section body`) tuple.
-        `section headers` is iterable of (`header name`, `header value`)
-        tuple.  `section value` is bytes, which is written directly.
+        :param headers: the iterable of (`name`, `value`) tuple of the header.
+        :param body: bytes of section body.
         """
         self.writer.write(SECTION_DELIMITER)
         for name, value in headers:
             self.header_writer.write(name, value)
-        if body:
-            self.writer.write(SECTION_BODY_START)
-            self.writer.write(body)
-            self.writer.write(SECTION_BODY_END)
+        self.writer.write(SECTION_BODY_START)
+        self.writer.write(body)
+        self.writer.write(SECTION_BODY_END)
