@@ -40,33 +40,23 @@ class AsyncNotifier(gntplib.Notifier):
                                   AsyncGNTPClient, io_loop=io_loop, **kwargs)
 
 
-class AsyncGNTPConnection(gntplib.GNTPConnectionCallbacksMixin):
+class AsyncGNTPConnection(gntplib.BaseGNTPConnection):
     """Asynchronous GNTP connection built on Tornado."""
 
     def __init__(self, address, timeout, final_callback, socket_callback=None,
                  io_loop=None):
+        gntplib.BaseGNTPConnection.__init__(self, final_callback,
+                                            socket_callback)
         sock = socket.create_connection(address, timeout=timeout)
         self.stream = iostream.IOStream(sock, io_loop=io_loop)
-        self.final_callback = final_callback
-        self.socket_callback = socket_callback
-        self.io_loop = io_loop
 
-    @gen.engine
-    def write_message(self, request):
-        """Send the request to the GNTP server.  If the request contains
-        :class:`AsyncIcon` instances, the request will be sent the GNTP server
-        after all their fetches are completed asynchronously."""
-        async_icons = collect_async_icons(request)
-        if async_icons:
-            yield gen.Task(fetch_async_icons_in_parallel,
-                           async_icons,
-                           self.io_loop)
-        self.stream.write(request.message)
+    def write_message(self, message):
+        """Send the request to the GNTP server."""
+        self.stream.write(message)
 
     def read_message(self, callback):
         """Read a message from opened stream and run callback with it."""
-        delimiter = gntplib.LINE_DELIMITER + gntplib.MESSAGE_DELIMITER
-        self.stream.read_until(delimiter, callback)
+        self.stream.read_until(gntplib.MESSAGE_DELIMITER, callback)
 
     def close(self):
         """Close the stream."""
@@ -77,21 +67,28 @@ class AsyncGNTPConnection(gntplib.GNTPConnectionCallbacksMixin):
 class AsyncGNTPClient(gntplib.GNTPClient):
     """Asynchronous GNTP client built on Tornado.
 
-    :param host: host of GNTP server.  Defaults to ``'localhost'``.
-    :param port: port of GNTP server.  Defaults to ``23053``.
-    :param timeout: timeout.  Defaults to ``10``.
     :param io_loop: :class:`tornado.ioloop.IOLoop` instance.
     """
 
-    connection_class = AsyncGNTPConnection
-
-    def __init__(self, host='localhost', port=23053, timeout=10, io_loop=None):
-        gntplib.GNTPClient.__init__(self, host, port, timeout)
+    def __init__(self, io_loop=None, **kwargs):
+        gntplib.GNTPClient.__init__(self, connection_class=AsyncGNTPConnection,
+                                    **kwargs)
         self.io_loop = io_loop
 
-    def _connect(self, final_callback, **kwargs):
-        """Connect to the GNTP server and return the connection."""
-        return gntplib.GNTPClient._connect(self, final_callback,
+    @gen.engine
+    def process_request(self, request, callback, **kwargs):
+        """Process a request.
+
+        If the request contains :class:`AsyncIcon` instances, the connection
+        to the GNTP server will be established after all their fetches are
+        completed asynchronously.
+        """
+        async_icons = collect_async_icons(request)
+        if async_icons:
+            yield gen.Task(fetch_async_icons_in_parallel,
+                           async_icons,
+                           self.io_loop)
+        gntplib.GNTPClient.process_request(self, request, callback,
                                            io_loop=self.io_loop, **kwargs)
 
 
